#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import base64
import random
import time
import unittest
from io import StringIO

import mock
import pandas as pd
from mock import call, MagicMock
from munch import Munch
from pandas.util.testing import assert_frame_equal

from neptune.exceptions import NoExperimentContext
from neptune.experiments import Experiment, push_new_experiment, get_current_experiment, pop_stopped_experiment
from neptune.internal.channels.channels import ChannelType, ChannelValue
from tests.neptune.random_utils import sort_df_by_columns, a_string, a_uuid_string, an_experiment_id, \
    a_project_qualified_name


class TestExperiment(unittest.TestCase):

    @mock.patch('neptune.experiments.ChannelsValuesSender', return_value=mock.MagicMock())
    @mock.patch('neptune.experiments.ExecutionContext', new=mock.MagicMock)
    def test_send_metric(self, ChannelsValuesSender):
        # given
        channels_values_sender = ChannelsValuesSender.return_value
        experiment = Experiment(mock.MagicMock(), an_experiment_id(), a_uuid_string(), a_project_qualified_name())
        channel_value = ChannelValue(
            x=random.randint(0, 100),
            y=dict(numeric_value=random.randint(0, 100)),
            ts=time.time()
        )

        # when
        experiment.send_metric('loss', channel_value.x, channel_value.y['numeric_value'], channel_value.ts)

        # then
        channels_values_sender.send.assert_called_with('loss', ChannelType.NUMERIC.value, channel_value)

    @mock.patch('neptune.experiments.ChannelsValuesSender', return_value=mock.MagicMock())
    @mock.patch('neptune.experiments.ExecutionContext', new=mock.MagicMock)
    def test_send_text(self, ChannelsValuesSender):
        # given
        channels_values_sender = ChannelsValuesSender.return_value
        experiment = Experiment(mock.MagicMock(), an_experiment_id(), a_uuid_string(), a_project_qualified_name())
        channel_value = ChannelValue(
            x=random.randint(0, 100),
            y=dict(text_value=a_string()),
            ts=time.time()
        )

        # when
        experiment.send_text('stdout', channel_value.x, channel_value.y['text_value'], channel_value.ts)

        # then
        channels_values_sender.send.assert_called_with('stdout', ChannelType.TEXT.value, channel_value)

    @mock.patch('neptune.experiments.get_image_content', return_value=b'content')
    @mock.patch('neptune.experiments.ChannelsValuesSender', return_value=mock.MagicMock())
    @mock.patch('neptune.experiments.ExecutionContext', new=mock.MagicMock)
    def test_send_image(self, ChannelsValuesSender, content):
        # given
        channels_values_sender = ChannelsValuesSender.return_value
        experiment = Experiment(mock.MagicMock(), an_experiment_id(), a_uuid_string(), a_project_qualified_name())
        image_value = dict(
            name=a_string(),
            description=a_string(),
            data=base64.b64encode(content()).decode('utf-8')
        )
        channel_value = ChannelValue(
            x=random.randint(0, 100),
            y=dict(image_value=image_value),
            ts=time.time()
        )

        # when
        experiment.send_image(
            'errors',
            channel_value.x,
            '/tmp/img.png',
            image_value['name'],
            image_value['description'],
            channel_value.ts
        )

        # then
        channels_values_sender.send.assert_called_with('errors', ChannelType.IMAGE.value, channel_value)

    def test_append_tags(self):
        # given
        client = mock.MagicMock()
        experiment = Experiment(client, an_experiment_id(), a_uuid_string(), a_project_qualified_name())

        #and
        def build_call(tags_list):
            return call(
                experiment=experiment,
                tags_to_add=tags_list,
                tags_to_delete=[]
            )

        # when
        experiment.append_tag('tag')
        experiment.append_tag(['tag1', 'tag2', 'tag3'])
        experiment.append_tag('tag1', 'tag2', 'tag3')
        experiment.append_tags('tag')
        experiment.append_tags(['tag1', 'tag2', 'tag3'])
        experiment.append_tags('tag1', 'tag2', 'tag3')

        # then
        client.update_tags.assert_has_calls([
            build_call(['tag']),
            build_call(['tag1', 'tag2', 'tag3']),
            build_call(['tag1', 'tag2', 'tag3']),
            build_call(['tag']),
            build_call(['tag1', 'tag2', 'tag3']),
            build_call(['tag1', 'tag2', 'tag3'])
        ])

    def test_get_numeric_channels_values(self):
        # when
        client = MagicMock()
        client.get_channel_points_csv.return_value = StringIO(u'\n'.join(['0.3,2.5', '1,2']))

        experiment = MagicMock()
        experiment.id = a_string()
        experiment.internal_id = a_uuid_string()
        experiment.channels = [Munch(id=a_uuid_string(), name='epoch_loss')]
        experiment.channelsLastValues = [Munch(channelName='epoch_loss', x=2.5, y=2)]

        client.get_experiment.return_value = experiment

        # then
        experiment = Experiment(
            client=client,
            _id=a_string(),
            internal_id=a_uuid_string(),
            project_full_id="test/sandbox"
        )
        result = experiment.get_numeric_channels_values('epoch_loss')

        expected_result = pd.DataFrame({'x': [0.3, 1.0],
                                        'epoch_loss': [2.5, 2.0]}, dtype=float)

        expected_result = sort_df_by_columns(expected_result)
        result = sort_df_by_columns(result)

        assert_frame_equal(expected_result, result)

    def test_get_current_experiment_from_stack(self):
        # given
        experiment = Munch(internal_id=a_uuid_string())

        # when
        push_new_experiment(experiment)

        # then
        self.assertEqual(get_current_experiment(), experiment)

    def test_pop_experiment_from_stack(self):
        # given
        first_experiment = Munch(internal_id=a_uuid_string())
        second_experiment = Munch(internal_id=a_uuid_string())
        # and
        push_new_experiment(first_experiment)

        # when
        push_new_experiment(second_experiment)

        # then
        self.assertEqual(get_current_experiment(), second_experiment)
        # and
        self.assertEqual(pop_stopped_experiment(), second_experiment)
        # and
        self.assertEqual(get_current_experiment(), first_experiment)

    def test_empty_stack(self):
        # when
        self.assertIsNone(pop_stopped_experiment())
        # and
        with self.assertRaises(NoExperimentContext):
            get_current_experiment()


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click
import pkg_resources


@click.group()
def main():
    pass


plugins = {
    entry_point.name: entry_point
    for entry_point
    in pkg_resources.iter_entry_points('neptune.plugins')
}

for name, entry_point in plugins.items():
    main.add_command(entry_point.load(), name)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sys
import time
import traceback

from neptune.internal.abort import DefaultAbortImpl, CustomAbortImpl
from neptune.internal.hardware.gauges.gauge_mode import GaugeMode
from neptune.internal.hardware.metrics.service.metric_service_factory import MetricServiceFactory
from neptune.internal.hardware.system.system_monitor import SystemMonitor
from neptune.internal.streams.stdstream_uploader import StdOutWithUpload, StdErrWithUpload
from neptune.internal.threads.aborting_thread import AbortingThread
from neptune.internal.threads.hardware_metric_reporting_thread import HardwareMetricReportingThread
from neptune.internal.threads.ping_thread import PingThread
from neptune.internal.websockets.reconnecting_websocket_factory import ReconnectingWebsocketFactory
from neptune.utils import is_notebook, in_docker


class ExecutionContext(object):

    def __init__(self, client, experiment):
        self._client = client
        self._experiment = experiment
        self._ping_thread = None
        self._hardware_metric_thread = None
        self._aborting_thread = None
        self._stdout_uploader = None
        self._stderr_uploader = None
        self._uncaught_exception_handler = sys.__excepthook__

        self._previous_uncaught_exception_handler = None

    def start(self,
              abort_callback=None,
              upload_stdout=True,
              upload_stderr=True,
              send_hardware_metrics=True,
              run_monitoring_thread=True,
              handle_uncaught_exceptions=True):

        if handle_uncaught_exceptions:
            self._set_uncaught_exception_handler()

        abortable = abort_callback is not None or DefaultAbortImpl.requirements_installed()

        if abortable:
            self._run_aborting_thread(abort_callback)

        if upload_stdout and not is_notebook():
            self._stdout_uploader = StdOutWithUpload(self._experiment)

        if upload_stderr and not is_notebook():
            self._stderr_uploader = StdErrWithUpload(self._experiment)

        if run_monitoring_thread:
            self._run_monitoring_thread()

        if send_hardware_metrics and SystemMonitor.requirements_installed():
            self._run_hardware_metrics_reporting_thread()

    def stop(self):
        if self._ping_thread:
            self._ping_thread.interrupt()
            self._ping_thread = None

        if self._hardware_metric_thread:
            self._hardware_metric_thread.interrupt()
            self._hardware_metric_thread = None

        if self._aborting_thread:
            self._aborting_thread.interrupt()
            self._aborting_thread = None

        if self._stdout_uploader:
            self._stdout_uploader.close()

        if self._stderr_uploader:
            self._stderr_uploader.close()

        sys.excepthook = self._previous_uncaught_exception_handler

    def _set_uncaught_exception_handler(self):

        def exception_handler(exc_type, exc_val, exc_tb):
            self._experiment.stop("\n".join(traceback.format_tb(exc_tb)) + "\n" + repr(exc_val))

            sys.__excepthook__(exc_type, exc_val, exc_tb)

        self._uncaught_exception_handler = exception_handler

        self._previous_uncaught_exception_handler = sys.excepthook
        sys.excepthook = exception_handler

    def _run_aborting_thread(self, abort_callback):
        if abort_callback is not None:
            abort_impl = CustomAbortImpl(abort_callback)
        else:
            abort_impl = DefaultAbortImpl(pid=os.getpid())

        websocket_factory = ReconnectingWebsocketFactory(
            client=self._client,
            experiment_id=self._experiment.internal_id
        )
        self._aborting_thread = AbortingThread(
            websocket_factory=websocket_factory,
            abort_impl=abort_impl,
            experiment_id=self._experiment.internal_id
        )
        self._aborting_thread.start()

    def _run_monitoring_thread(self):
        self._ping_thread = PingThread(client=self._client, experiment=self._experiment)
        self._ping_thread.start()

    def _run_hardware_metrics_reporting_thread(self):
        gauge_mode = GaugeMode.CGROUP if in_docker() else GaugeMode.SYSTEM
        metric_service = MetricServiceFactory(self._client, os.environ).create(
            gauge_mode=gauge_mode,
            experiment=self._experiment,
            reference_timestamp=time.time()
        )

        self._hardware_metric_thread = HardwareMetricReportingThread(
            metric_service=metric_service,
            metric_sending_interval_seconds=3
        )
        self._hardware_metric_thread.start()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from platform import node as get_hostname

import click

import pandas as pd

from neptune.experiments import Experiment, push_new_experiment
from neptune.internal.abort import DefaultAbortImpl
from neptune.utils import as_list, map_keys, get_git_info, discover_git_repo_location


class Project(object):
    # pylint: disable=redefined-builtin

    """It contains all the information about a Neptune project

    You can extract the experiment view in a form of a dataframe or a list of experiments.
    Project lets you do filtering based on conditions not to fetch the entire, sometimes huge list of experiments.

    Args:
        client(`neptune.Client`): Client object
        internal_id:
        namespace(str): It can either be your organization or user name. You can list all the public projects for any
               organization or user you want as long as you know their namespace.
        name(str): short project name.

    Attributes:
        client(`neptune.Client`): Client object
        internal_id:
        namespace(str): It can either be your organization or user name. You can list all the public projects for any
               organization or user you want as long as you know their namespace.
        name(str): short project name.

    Examples:
        Instantiate a session.

        >>> from neptune.sessions import Session
        >>> session = Session()

        Fetch a project.

        >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']
        >>> project
        Project(neptune-ml/Salt-Detection)

    Todo:
        Drop the pylint line.
        Explain what internal_id is

    """

    def __init__(self, client, internal_id, namespace, name):
        self.client = client
        self.internal_id = internal_id
        self.namespace = namespace
        self.name = name

    def get_members(self):
        """Retrieve a list of project members.

        Returns:
            list: A list of usernames of project members.

        Examples:

            >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']
            >>> project.get_members()

        """
        project_members = self.client.get_project_members(self.internal_id)
        return [member.registeredMemberInfo.username for member in project_members if member.registeredMemberInfo]

    def get_experiments(self, id=None, state=None, owner=None, tag=None, min_running_time=None):
        """Retrieve a list of experiments matching the specified criteria.

        All of the parameters of this method are optional, each of them specifies a single criterion.

        Only experiments matching all of the criteria will be returned.

        If a specific criterion accepts a list (like `state`), experiments matching any element of the list
        match this criterion.

        Args:
            id(list): An ID or list of experiment IDs (rowo.g. 'SAN-1' or ['SAN-1', 'SAN-2'])
            state(list): A state or list of experiment states.
                E.g. 'succeeded' or ['succeeded', 'preempted'].
                Possible states: 'creating', 'waiting', 'initializing', 'running', 'cleaning',
                'crashed', 'failed', 'aborted', 'preempted', 'succeeded'
            owner(list): The owner or list of owners of the experiments. This parameter expects usernames.
            tag(list): A tag or a list of experiment tags. E.g. 'solution-1' or ['solution-1', 'solution-2'].
            min_running_time(int): Minimum running time of an experiment in seconds.

        Returns:
            list: List of `Experiment` objects

        Examples:
            Instantiate a session.

            >>> from neptune.sessions import Session
            >>> session = Session()

            Fetch a project.

            >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']

            Finally, get a list of experiments that satisfies your criteria:

            >>> project.get_experiments(state=['aborted'], owner=['neyo'], min_running_time=100000)
            [Experiment(SAL-1609),
             Experiment(SAL-1765),
             Experiment(SAL-1941),
             Experiment(SAL-1960),
             Experiment(SAL-2025)]

        """
        leaderboard_entries = self._fetch_leaderboard(id, state, owner, tag, min_running_time)
        return [
            Experiment(self.client, entry.id, entry.internal_id, entry.project_full_id) for entry in leaderboard_entries
        ]

    def get_leaderboard(self, id=None, state=None, owner=None, tag=None, min_running_time=None):
        """Fetches Neptune experiment view to pandas DataFrame

        Retrieve experiments matching the specified criteria and present them in a form of a DataFrame
        resembling Neptune's leaderboard.

        The returned DataFrame contains columns for all system properties,
        numeric and text channels, user-defined properties and parameters defined
        in the selected experiments (not across the entire project).

        Every row in this DataFrame represents a single experiment. As such, some columns may be empty,
        since experiments define various channels, properties, etc.

        For each channel at most one (the last one) value is returned per experiment.
        Text values are trimmed to 255 characters.

        All of the parameters of this method are optional, each of them specifies a single criterion.

        Only experiments matching all of the criteria will be returned.

        If a specific criterion accepts a list (like `state`), experiments matching any element of the list
        match this criterion.

        Args:
            id(list): An ID or list of experiment IDs (rowo.g. 'SAN-1' or ['SAN-1', 'SAN-2'])
            state(list): A state or list of experiment states.
                E.g. 'succeeded' or ['succeeded', 'preempted']
                Possible states: 'creating', 'waiting', 'initializing', 'running',
                    'cleaning', 'crashed', 'failed', 'aborted', 'preempted', 'succeeded'
            owner(list): The owner or list of owners of the experiments. This parameter expects usernames.
            tag(list): A tag or a list of experiment tags. E.g. 'solution-1' or ['solution-1', 'solution-2'].
            min_running_time(int): Minimum running time of an experiment in seconds.

        Returns:
            `pandas.DataFrame`: Neptune experiment view in the form of a dataframe.

        Examples:
            Instantiate a session.

            >>> from neptune.sessions import Session
            >>> session = Session()

            Fetch a project.

            >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']

            Finally, get a dataframe that resembles experiment view. It is constructed from all the
            experiments that satisfy your criteria:

            >>> project.get_leaderboard(state=['aborted'], owner=['neyo'], min_running_time=100000)

        Todo:
            tags - is it ok now?
        """

        leaderboard_entries = self._fetch_leaderboard(id, state, owner, tag, min_running_time)

        def make_row(entry):
            channels = dict(
                ('channel_{}'.format(ch.name), ch.trimmed_y) for ch in entry.channels
            )

            parameters = map_keys('parameter_{}'.format, entry.parameters)
            properties = map_keys('property_{}'.format, entry.properties)

            r = {}
            r.update(entry.system_properties)
            r.update(channels)
            r.update(parameters)
            r.update(properties)
            return r

        rows = ((n, make_row(e)) for (n, e) in enumerate(leaderboard_entries))

        df = pd.DataFrame.from_dict(data=dict(rows), orient='index')
        df = df.reindex(self._sort_leaderboard_columns(df.columns), axis='columns')
        return df

    def create_experiment(self,
                          name=None,
                          description=None,
                          params=None,
                          properties=None,
                          tags=None,
                          upload_source_files=None,
                          abort_callback=None,
                          upload_stdout=True,
                          upload_stderr=True,
                          send_hardware_metrics=True,
                          run_monitoring_thread=True,
                          handle_uncaught_exceptions=True,
                          git_info=None,
                          hostname=None):
        """
        Raises:
            `ExperimentValidationError`: When provided arguments are invalid.
            `ExperimentLimitReached`: When experiment limit in the project has been reached.
        """

        if name is None:
            name = "Untitled"

        if description is None:
            description = ""

        if params is None:
            params = {}

        if properties is None:
            properties = {}

        if tags is None:
            tags = []

        if git_info is None:
            git_info = get_git_info(discover_git_repo_location())

        if hostname is None:
            hostname = get_hostname()

        abortable = abort_callback is not None or DefaultAbortImpl.requirements_installed()

        experiment = self.client.create_experiment(
            project=self,
            name=name,
            description=description,
            params=params,
            properties=properties,
            tags=tags,
            abortable=abortable,
            monitored=run_monitoring_thread,
            git_info=git_info,
            hostname=hostname
        )

        experiment.start(
            upload_source_files=upload_source_files,
            abort_callback=abort_callback,
            upload_stdout=upload_stdout,
            upload_stderr=upload_stderr,
            send_hardware_metrics=send_hardware_metrics,
            run_monitoring_thread=run_monitoring_thread,
            handle_uncaught_exceptions=handle_uncaught_exceptions
        )

        push_new_experiment(experiment)

        click.echo(str(experiment.id))
        click.echo(self._get_experiment_link(experiment))

        return experiment

    def _get_experiment_link(self, experiment):
        return "{base_url}/{namespace}/{project}/e/{exp_id}".format(
            base_url=self.client.api_address,
            namespace=self.namespace,
            project=self.name,
            exp_id=experiment.id
        )

    def create_notebook(self):
        notebook = self.client.create_notebook(self)
        return notebook

    def get_notebook(self, notebook_id):
        return self.client.get_notebook(project=self, notebook_id=notebook_id)

    @property
    def full_id(self):
        """Creates a full project id by combining the namespace and project name.
        """
        return '{}/{}'.format(self.namespace, self.name)

    def __str__(self):
        return 'Project({})'.format(self.full_id)

    def __repr__(self):
        return str(self)

    def __eq__(self, o):
        return self.__dict__ == o.__dict__

    def __ne__(self, o):
        return not self.__eq__(o)

    def _fetch_leaderboard(self, id, state, owner, tag, min_running_time):
        return self.client.get_leaderboard_entries(
            project=self, ids=as_list(id), states=as_list(state),
            owners=as_list(owner), tags=as_list(tag),
            min_running_time=min_running_time)

    @staticmethod
    def _sort_leaderboard_columns(column_names):
        user_defined_weights = {
            'channel': 1,
            'parameter': 2,
            'property': 3
        }

        system_properties_weights = {
            'id': 0,
            'name': 1,
            'created': 2,
            'finished': 3,
            'owner': 4,
            'worker_type': 5,
            'environment': 6,
        }

        def key(c):
            """A sorting key for a column name.

            Sorts by the system properties first, then channels, parameters, user-defined properties.

            Within each group columns are sorted alphabetically, except for system properties,
            where order is custom.
            """
            parts = c.split('_', 1)
            if parts[0] in user_defined_weights.keys():
                name = parts[1]
                weight = user_defined_weights.get(parts[0], 99)
                system_property_weight = None
            else:
                name = c
                weight = 0
                system_property_weight = system_properties_weights.get(name, 99)

            return weight, system_property_weight, name

        return sorted(column_names, key=key)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re

from neptune.internal.websockets.reconnecting_websocket import ReconnectingWebsocket


class ReconnectingWebsocketFactory(object):
    def __init__(self, client, experiment_id):
        self._client = client
        self._base_address = re.sub(r'^http', 'ws', self._client.api_address) + '/api/notifications/v1'
        self._experiment_id = experiment_id

    def create(self, shutdown_condition):
        url = self._experiment_url(self._base_address, self._experiment_id)
        return ReconnectingWebsocket(
            url=url, oauth2_session=self._client.authenticator.auth.session, shutdown_event=shutdown_condition)

    @staticmethod
    def _experiment_url(base_address, experiment_id):
        return base_address + '/experiments/' + experiment_id + '/operations'



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

from neptune.utils import validate_notebook_path


class Notebook(object):
    """It contains all the information about a Neptune Notebook

    Args:
        client(`neptune.Client`): Client object
        project(`neptune.Project`):
        _id(`str`): Notebook uuid
        owner(`str`):


    Examples:
        Instantiate a session and fetch a project.

        >>> import neptune
        >>> project = neptune.init()

        Create a notebook.

        >>> notebook = project.create_notebook('file.ipynb')
    """

    def __init__(self, client, project, _id, owner):
        self._client = client
        self._project = project
        self._id = _id
        self._owner = owner

    @property
    def id(self):
        return self._id

    @property
    def owner(self):
        return self._owner

    def add_checkpoint(self, file_path):
        validate_notebook_path(file_path)

        with open(file_path) as f:
            return self._client.create_checkpoint(self.id, os.path.abspath(file_path), f)

    def get_path(self):
        return self._client.get_last_checkpoint(self._project, self._id).path

    def get_name(self):
        return self._client.get_last_checkpoint(self._project, self._id).name



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import base64
import os
import sys
import threading
import traceback

import pandas as pd
import six
from pandas.errors import EmptyDataError

from neptune.api_exceptions import ExperimentAlreadyFinished
from neptune.exceptions import FileNotFound, InvalidChannelValue, NoChannelValue, NoExperimentContext
from neptune.internal.channels.channels import ChannelValue, ChannelType, ChannelNamespace
from neptune.internal.channels.channels_values_sender import ChannelsValuesSender
from neptune.internal.execution.execution_context import ExecutionContext
from neptune.internal.storage.storage_utils import upload_to_storage
from neptune.internal.utils.image import get_image_content
from neptune.utils import align_channels_on_x, is_float


class Experiment(object):
    """It contains all the information about a Neptune Experiment

    This class lets you extract experiment by, short experiment id, names of all the channels,
    system properties and other properties, parameters, numerical channel values,
    information about the hardware utilization during the experiment

    Args:
        client(`neptune.Client'): Client object
        _id(`str`)
        internal_id(`str`): UUID
        project_full_id(`str`)


    Examples:
        Instantiate a session.

        >>> from neptune.sessions import Session
        >>> session = Session()

        Fetch a project and a list of experiments.

        >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']
        >>> experiments = project.get_experiments(state=['aborted'], owner=['neyo'], min_running_time=100000)

        Get an experiment instance.

        >>> experiment = experiments[0]
        >>> experiment
        Experiment(SAL-1609)

    Todo:
        Column sorting
    """

    def __init__(self, client, _id, internal_id, project_full_id):
        self._client = client
        self._id = _id
        self._internal_id = internal_id
        self._project_full_id = project_full_id
        self._channels_values_sender = ChannelsValuesSender(self)
        self._execution_context = ExecutionContext(client, self)

    @property
    def id(self):
        """ Experiment short id

        Examples:
            Instantiate a session.

            >>> from neptune.sessions import Session
            >>> session = Session()

            Fetch a project and a list of experiments.

            >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']
            >>> experiments = project.get_experiments(state=['aborted'], owner=['neyo'], min_running_time=100000)

            Get an experiment instance.

            >>> experiment = experiments[0]

            Get experiment short id.

            >>> experiment.id
            'SAL-1609'

        """
        return self._id

    @property
    def name(self):
        return self._client.get_experiment(self._internal_id).name

    @property
    def state(self):
        return self._client.get_experiment(self._internal_id).state

    @property
    def internal_id(self):
        return self._internal_id

    @property
    def limits(self):
        # TODO: get limit from server
        return {
            'channels': {
                'numeric': 1000,
                'text': 100,
                'image': 100
            }
        }

    def get_system_properties(self):
        """Retrieve system properties like owner, times of creation and completion, worker type, etc.

        Returns:
            dict: A dictionary mapping a property name to value.

        Examples:
            Instantiate a session.

            >>> from neptune.sessions import Session
            >>> session = Session()

            Fetch a project and a list of experiments.

            >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']
            >>> experiments = project.get_experiments(state=['aborted'], owner=['neyo'], min_running_time=100000)

            Get an experiment instance.

            >>> experiment = experiments[0]

            Get experiment system properties.

            >>> experiment.get_system_properties

        Note:
            The list of supported system properties may change over time.

        """
        experiment = self._client.get_experiment(self._internal_id)
        return {
            'id': experiment.shortId,
            'name': experiment.name,
            'created': experiment.timeOfCreation,
            'finished': experiment.timeOfCompletion,
            'running_time': experiment.runningTime,
            'owner': experiment.owner,
            'size': experiment.storageSize,
            'tags': experiment.tags,
            'notes': experiment.description
        }

    def get_tags(self):
        return self._client.get_experiment(self._internal_id).tags

    def append_tag(self, tag, *tags):
        if isinstance(tag, list):
            tags_list = tag
        else:
            tags_list = [tag] + list(tags)
        self._client.update_tags(experiment=self,
                                 tags_to_add=tags_list,
                                 tags_to_delete=[])

    def append_tags(self, tag, *tags):
        self.append_tag(tag, *tags)

    def remove_tag(self, tag):
        self._client.update_tags(experiment=self,
                                 tags_to_add=[],
                                 tags_to_delete=[tag])

    def get_channels(self):
        """Retrieve all channel names along with their representations for this experiment.

        Returns:
            dict: A dictionary mapping a channel name to channel.

        Examples:
            Instantiate a session.

            >>> from neptune.sessions import Session
            >>> session = Session()

            Fetch a project and a list of experiments.

            >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']
            >>> experiments = project.get_experiments(state=['aborted'], owner=['neyo'], min_running_time=100000)

            Get an experiment instance.

            >>> experiment = experiments[0]

            Get experiment channels.

            >>> experiment.get_channels()

        """
        experiment = self._client.get_experiment(self.internal_id)
        channels_last_values_by_name = dict((ch.channelName, ch) for ch in experiment.channelsLastValues)
        channels = dict()
        for ch in experiment.channels:
            last_value = channels_last_values_by_name.get(ch.name, None)
            if last_value is not None:
                ch.x = last_value.x
                ch.y = last_value.y
            elif ch.lastX is not None:
                ch.x = ch.lastX
                ch.y = None
            else:
                ch.x = None
                ch.y = None
            channels[ch.name] = ch
        return channels

    def get_system_channels(self):
        """Retrieve all system channel names along with their representations for this experiment.

        Returns:
            dict: A dictionary mapping a channel name to channel.

        Examples:
            Instantiate a session.

            >>> from neptune.sessions import Session
            >>> session = Session()

            Fetch a project and a list of experiments.

            >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']
            >>> experiments = project.get_experiments(state=['aborted'], owner=['neyo'], min_running_time=100000)

            Get an experiment instance.

            >>> experiment = experiments[0]

            Get experiment channels.

            >>> experiment.get_system_channels()

        """
        channels = self._client.get_system_channels(self)
        return dict((ch.name, ch) for ch in channels)

    def upload_source_files(self, source_files):
        """
        Raises:
            `StorageLimitReached`: When storage limit in the project has been reached.
        """
        files_list = []
        for source_file in source_files:
            if not os.path.exists(source_file):
                raise FileNotFound(source_file)
            files_list.append((os.path.abspath(source_file), source_file))

        upload_to_storage(files_list=files_list,
                          upload_api_fun=self._client.upload_experiment_source,
                          upload_tar_api_fun=self._client.extract_experiment_source,
                          experiment=self)

    def send_metric(self, channel_name, x, y=None, timestamp=None):
        x, y = self._get_valid_x_y(x, y)

        if not is_float(y):
            raise InvalidChannelValue(expected_type='float', actual_type=type(y).__name__)

        value = ChannelValue(x, dict(numeric_value=y), timestamp)
        self._channels_values_sender.send(channel_name, ChannelType.NUMERIC.value, value)

    def send_text(self, channel_name, x, y=None, timestamp=None):
        x, y = self._get_valid_x_y(x, y)

        if not isinstance(y, six.string_types):
            raise InvalidChannelValue(expected_type='str', actual_type=type(y).__name__)

        value = ChannelValue(x, dict(text_value=y), timestamp)
        self._channels_values_sender.send(channel_name, ChannelType.TEXT.value, value)

    def send_image(self, channel_name, x, y=None, name=None, description=None, timestamp=None):
        x, y = self._get_valid_x_y(x, y)

        input_image = dict(
            name=name,
            description=description,
            data=base64.b64encode(get_image_content(y)).decode('utf-8')
        )

        value = ChannelValue(x, dict(image_value=input_image), timestamp)
        self._channels_values_sender.send(channel_name, ChannelType.IMAGE.value, value)

    def send_artifact(self, artifact):
        """
        Raises:
            `StorageLimitReached`: When storage limit in the project has been reached.
        """
        if not os.path.exists(artifact):
            raise FileNotFound(artifact)

        upload_to_storage(files_list=[(os.path.abspath(artifact), os.path.basename(artifact))],
                          upload_api_fun=self._client.upload_experiment_output,
                          upload_tar_api_fun=self._client.extract_experiment_output,
                          experiment=self)

    def send_graph(self, graph_id, value):
        """Upload a tensorflow graph for this experiment.

        Args:
            graph_id: a string UUID identifying the graph (managed by user)
            value: a string representation of Tensorflow graph

        Examples:
            Instantiate a session.

            >>> from neptune.sessions import Session
            >>> session = Session()

            Fetch a project and a list of experiments.

            >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']
            >>> experiments = project.get_experiments(state=['aborted'], owner=['neyo'], min_running_time=100000)

            Get an experiment instance.

            >>> experiment = experiments[0]

            Send graph to experiment.
            >>> import uuid
            >>> experiment.send_graph(str(uuid.uuid4()), str("tf.GraphDef instance"))

        """

        self._client.put_tensorflow_graph(self, graph_id, value)

    def get_parameters(self):
        """Retrieve parameters for this experiment.

        Returns:
            dict: A dictionary mapping a parameter name to value.

        Examples:
            Instantiate a session.

            >>> from neptune.sessions import Session
            >>> session = Session()

            Fetch a project and a list of experiments.

            >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']
            >>> experiments = project.get_experiments(state=['aborted'], owner=['neyo'], min_running_time=100000)

            Get an experiment instance.

            >>> experiment = experiments[0]

            Get experiment parameters.

            >>> experiment.get_parameters()

        """
        experiment = self._client.get_experiment(self.internal_id)
        return dict((p.name, self._convert_parameter_value(p.value, p.parameterType)) for p in experiment.parameters)

    def get_properties(self):
        """Retrieve user-defined properties for this experiment.

        Returns:
            dict: A dictionary mapping a property key to value.

        Examples:
            Instantiate a session.

            >>> from neptune.sessions import Session
            >>> session = Session()

            Fetch a project and a list of experiments.

            >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']
            >>> experiments = project.get_experiments(state=['aborted'], owner=['neyo'], min_running_time=100000)

            Get an experiment instance.

            >>> experiment = experiments[0]

            Get experiment properties.

            >>> experiment.get_properties

        """
        experiment = self._client.get_experiment(self.internal_id)
        return dict((p.key, p.value) for p in experiment.properties)

    def set_property(self, key, value):
        properties = {p.key: p.value for p in self._client.get_experiment(self.internal_id).properties}
        properties[key] = value
        return self._client.update_experiment(
            experiment=self,
            properties=properties
        )

    def remove_property(self, key):
        properties = {p.key: p.value for p in self._client.get_experiment(self.internal_id).properties}
        del properties[key]
        return self._client.update_experiment(
            experiment=self,
            properties=properties
        )

    def get_hardware_utilization(self):
        """Retrieve RAM, CPU and GPU utilization throughout the experiment.

        The returned DataFrame contains 2 columns (x_*, y_*) for each of: RAM, CPU and each GPU.
        The x_ column contains the time (in milliseconds) from the experiment start,
        while the y_ column contains the value of the appropriate metric.

        RAM and GPU memory usage is returned in gigabytes.
        CPU and GPU utilization is returned as a percentage (0-100).

        E.g. For an experiment using a single GPU, this method will return a DataFrame
        of the following columns:

        x_ram, y_ram, x_cpu, y_cpu, x_gpu_util_0, y_gpu_util_0, x_gpu_mem_0, y_gpu_mem_0

        The following values denote that after 3 seconds, the experiment used 16.7 GB of RAM.
        x_ram, y_ram = 3000, 16.7

        The returned DataFrame may contain NaNs if one of the metrics has more values than others.

        Returns:
            `pandas.DataFrame`: Dataframe containing the hardware utilization metrics throughout the experiment.

        Examples:
            Instantiate a session.

            >>> from neptune.sessions import Session
            >>> session = Session()

            Fetch a project and a list of experiments.

            >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']
            >>> experiments = project.get_experiments(state=['aborted'], owner=['neyo'], min_running_time=100000)

            Get an experiment instance.

            >>> experiment = experiments[0]

            Get hardware utilization channels.

            >>> experiment.get_hardware_utilization

        """
        metrics_csv = self._client.get_metrics_csv(self)
        try:
            return pd.read_csv(metrics_csv)
        except EmptyDataError:
            return pd.DataFrame()

    def get_numeric_channels_values(self, *channel_names):
        """
        Retrieve values of specified numeric channels.

        The returned DataFrame contains 1 additional column x along with the requested channels.

        E.g. get_numeric_channels_values('loss', 'auc') will return a DataFrame of the following structure:
            x, loss, auc

        The returned DataFrame may contain NaNs if one of the channels has more values than others.

        Args:
            *channel_names: variable length list of names of the channels to retrieve values for.

        Returns:
            `pandas.DataFrame`: Dataframe containing the values for the requested numerical channels.

        Examples:
            Instantiate a session.

            >>> from neptune.sessions import Session
            >>> session = Session()

            Fetch a project and a list of experiments.

            >>> project = session.get_projects('neptune-ml')['neptune-ml/Salt-Detection']
            >>> experiments = project.get_experiments(state=['aborted'], owner=['neyo'], min_running_time=100000)

            Get an experiment instance.

            >>> exp = experiments[0]

            Get numeric channel value for channels 'unet_0 batch sum loss' and 'unet_1 batch sum loss'.

            >>> batch_channels = exp.get_numeric_channels_values('unet_0 batch sum loss', 'unet_1 batch sum loss')
            >>> epoch_channels = exp.get_numeric_channels_values('unet_0 epoch_val sum loss', 'Learning Rate')

        Note:
            Remember to fetch the dataframe for the channels that have a common temporal/iteration axis x.
            For example combine epoch channels to one dataframe and batch channels to the other
        """

        channels_data = {}
        channels_by_name = self.get_channels()
        for channel_name in channel_names:
            channel_id = channels_by_name[channel_name].id
            try:
                channels_data[channel_name] = pd.read_csv(
                    self._client.get_channel_points_csv(self, channel_id),
                    header=None,
                    names=['x_{}'.format(channel_name), 'y_{}'.format(channel_name)],
                    dtype=float
                )
            except EmptyDataError:
                channels_data[channel_name] = pd.DataFrame(
                    columns=['x_{}'.format(channel_name), 'y_{}'.format(channel_name)],
                    dtype=float
                )

        return align_channels_on_x(pd.concat(channels_data.values(), axis=1, sort=False))

    def start(self,
              upload_source_files=None,
              abort_callback=None,
              upload_stdout=True,
              upload_stderr=True,
              send_hardware_metrics=True,
              run_monitoring_thread=True,
              handle_uncaught_exceptions=True):

        if upload_source_files is None:
            main_file = sys.argv[0]
            main_abs_path = os.path.join(os.getcwd(), os.path.basename(main_file))
            if os.path.isfile(main_abs_path):
                upload_source_files = [os.path.relpath(main_abs_path, os.getcwd())]
            else:
                upload_source_files = []

        self.upload_source_files(upload_source_files)

        self._execution_context.start(
            abort_callback=abort_callback,
            upload_stdout=upload_stdout,
            upload_stderr=upload_stderr,
            send_hardware_metrics=send_hardware_metrics,
            run_monitoring_thread=run_monitoring_thread,
            handle_uncaught_exceptions=handle_uncaught_exceptions
        )

    def stop(self, exc_tb=None):

        self._channels_values_sender.join()

        try:
            if exc_tb is None:
                self._client.mark_succeeded(self)
            else:
                self._client.mark_failed(self, exc_tb)
        except ExperimentAlreadyFinished:
            pass

        self._execution_context.stop()

        pop_stopped_experiment()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_tb is None:
            self.stop()
        else:
            self.stop("\n".join(traceback.format_tb(exc_tb)) + "\n" + repr(exc_val))

    def __str__(self):
        return 'Experiment({})'.format(self.id)

    def __repr__(self):
        return str(self)

    def __eq__(self, o):
        # pylint: disable=protected-access
        return self._id == o._id and self._internal_id == o._internal_id and self._project_full_id == o._project_full_id

    def __ne__(self, o):
        return not self.__eq__(o)

    @staticmethod
    def _convert_parameter_value(value, parameter_type):
        if parameter_type == 'double':
            return float(value)
        else:
            return value

    @staticmethod
    def _get_valid_x_y(x, y):
        if x is None:
            raise NoChannelValue()

        if y is None:
            y = x
            x = None
        elif not is_float(x):
            raise InvalidChannelValue(expected_type='float', actual_type=type(x).__name__)

        return x, y

    def _send_channels_values(self, channels_with_values):
        self._client.send_channels_values(self, channels_with_values)

    def _get_channels(self, channels_names_with_types):
        existing_channels = self.get_channels()
        channels_by_name = {}
        for (channel_name, channel_type) in channels_names_with_types:
            channel = existing_channels.get(channel_name, None)
            if channel is None:
                channel = self._create_channel(channel_name, channel_type)
            channels_by_name[channel.name] = channel
        return channels_by_name

    def _get_channel(self, channel_name, channel_type, channel_namespace=ChannelNamespace.USER):
        channel = self._find_channel(channel_name, channel_namespace)
        if channel is None:
            channel = self._create_channel(channel_name, channel_type, channel_namespace)
        return channel

    def _find_channel(self, channel_name, channel_namespace):
        if channel_namespace == ChannelNamespace.USER:
            return self.get_channels().get(channel_name, None)
        elif channel_namespace == ChannelNamespace.SYSTEM:
            return self.get_system_channels().get(channel_name, None)
        else:
            raise RuntimeError("Unknown channel namesapce {}".format(channel_namespace))

    def _create_channel(self, channel_name, channel_type, channel_namespace=ChannelNamespace.USER):
        if channel_namespace == ChannelNamespace.USER:
            return self._client.create_channel(self, channel_name, channel_type)
        elif channel_namespace == ChannelNamespace.SYSTEM:
            return self._client.create_system_channel(self, channel_name, channel_type)
        else:
            raise RuntimeError("Unknown channel namesapce {}".format(channel_namespace))


_experiments_stack = []

__lock = threading.RLock()


def get_current_experiment():
    # pylint: disable=global-statement
    global _experiments_stack
    with __lock:
        if _experiments_stack:
            return _experiments_stack[len(_experiments_stack) - 1]
        else:
            raise NoExperimentContext()


def push_new_experiment(new_experiment):
    # pylint: disable=global-statement
    global _experiments_stack, __lock
    with __lock:
        _experiments_stack.append(new_experiment)
        return new_experiment


def pop_stopped_experiment():
    # pylint: disable=global-statement
    global _experiments_stack, __lock
    with __lock:
        if _experiments_stack:
            stopped_experiment = _experiments_stack.pop()
        else:
            stopped_experiment = None
        return stopped_experiment



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import OrderedDict
import re

from neptune.api_exceptions import ProjectNotFound
from neptune.client import Client
from neptune.credentials import Credentials
from neptune.exceptions import IncorrectProjectQualifiedName
from neptune.patterns import PROJECT_QUALIFIED_NAME_PATTERN
from neptune.projects import Project


class Session(object):
    """It handles the communication with the Neptune experiment database.

    In order to query Neptune experiment database in any way you have to instantiate this object first.

    Args:
        api_token(str): This is a secret API key that you can retrieve by running
            `$ neptune account api-token get`.

    Attributes:
        credentials (`neptune.Credentials`): `Credentials` object instance that authenticates your
            calls to Neptune API.

    Examples:
        Examples should be written in doctest format, and should illustrate how
        to use the function.

        >>> from neptune.sessions import Session
        >>> session = Session(api_token='YOUR_NEPTUNE_API_TOKEN')

        or assuming you have created an environment variable by running:

        `export NEPTUNE_API_TOKEN=YOUR_NEPTUNE_API_TOKEN`

        and simply go:

        >>> session = Session()
    """

    def __init__(self, api_token=None, proxies=None):
        credentials = Credentials(api_token)

        self.credentials = credentials
        self.proxies = proxies

        self._client = Client(self.credentials.api_address, self.credentials.api_token, proxies)

    def get_project(self, project_qualified_name):
        """
        Raises:
            `ProjectNotFound`: When a project with given name does not exist.
        """
        if not project_qualified_name:
            raise ProjectNotFound(project_qualified_name)

        if not re.match(PROJECT_QUALIFIED_NAME_PATTERN, project_qualified_name):
            raise IncorrectProjectQualifiedName(project_qualified_name)

        project = self._client.get_project(project_qualified_name)
        return Project(
            client=self._client,
            internal_id=project.id,
            namespace=project.organizationName,
            name=project.name
        )

    def get_projects(self, namespace):
        """It gets all project and full project names for given namespace

        In order to access experiment data one needs to get a `Project` object first. This method helps you figure out
        what are the available projects and access the project of interest.
        You can list both your private and public projects.
        You can also access all the public projects that belong to any user or organization,
        as long as you know what is their namespace.

        Args:
            namespace(str): It can either be your organization or user name. You can list all the public projects
                for any organization or user you want as long as you know their namespace.

        Returns:
            dict: Dictionary of "NAMESPACE/PROJECT_NAME" and `neptune.project.Project` object pairs that contains
            all the projects that belong to the selected namespace.

        Raises:
            `NamespaceNotFound`: When the given namespace does not exist.

        Examples:
            First, you need to create a Session instance:

            >>> from neptune.sessions import Session
            >>> session = Session()

            Now, you can list all the projects available for a selected namespace. You can use `YOUR_NAMESPACE` which
            is your organization or user name. You can also list public projects created by other organizations.
            For example you can use the `neptune-ml` namespace.

            >>> session.get_projects('neptune-ml')
            {'neptune-ml/Sandbox': Project(neptune-ml/Sandbox),
            'neptune-ml/Home-Credit-Default-Risk': Project(neptune-ml/Home-Credit-Default-Risk),
            'neptune-ml/Mapping-Challenge': Project(neptune-ml/Mapping-Challenge),
            'neptune-ml/Ships': Project(neptune-ml/Ships),
            'neptune-ml/human-protein-atlas': Project(neptune-ml/human-protein-atlas),
            'neptune-ml/Salt-Detection': Project(neptune-ml/Salt-Detection),
            'neptune-ml/Data-Science-Bowl-2018': Project(neptune-ml/Data-Science-Bowl-2018)}
        """

        projects = [Project(self._client, p.id, namespace, p.name) for p in self._client.get_projects(namespace)]
        return OrderedDict((p.full_id, p) for p in projects)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import threading
import time
from collections import namedtuple
from itertools import groupby

from future.moves import queue

from neptune.api_exceptions import NeptuneApiException
from neptune.internal.channels.channels import ChannelIdWithValues, ChannelNameWithTypeAndNamespace, ChannelValue,\
    ChannelType, ChannelNamespace
from neptune.internal.threads.neptune_thread import NeptuneThread

_logger = logging.getLogger(__name__)


class ChannelsValuesSender(object):
    _QUEUED_CHANNEL_VALUE = namedtuple(
        "QueuedChannelValue",
        ['channel_name', 'channel_type', 'channel_value', 'channel_namespace']
    )

    __LOCK = threading.RLock()

    def __init__(self, experiment):
        self._experiment = experiment
        self._values_queue = None
        self._sending_thread = None

    def send(self, channel_name, channel_type, channel_value, channel_namespace=ChannelNamespace.USER):
        with self.__LOCK:
            if not self._is_running():
                self._start()

        self._values_queue.put(self._QUEUED_CHANNEL_VALUE(
            channel_name=channel_name,
            channel_type=channel_type,
            channel_value=channel_value,
            channel_namespace=channel_namespace
        ))

    def join(self):
        with self.__LOCK:
            if self._is_running():
                self._sending_thread.interrupt()
                self._sending_thread.join()
                self._sending_thread = None
                self._values_queue = None

    def _is_running(self):
        return self._values_queue is not None and self._sending_thread is not None and self._sending_thread.is_alive()

    def _start(self):
        self._values_queue = queue.Queue()
        self._sending_thread = ChannelsValuesSendingThread(self._experiment, self._values_queue)
        self._sending_thread.start()


class ChannelsValuesSendingThread(NeptuneThread):
    _SLEEP_TIME = 5
    _MAX_VALUES_BATCH_LENGTH = 100
    _MAX_IMAGE_VALUES_BATCH_LENGTH = 10

    def __init__(self, experiment, values_queue):
        super(ChannelsValuesSendingThread, self).__init__(is_daemon=False)
        self._values_queue = values_queue
        self._experiment = experiment
        self._sleep_time = self._SLEEP_TIME
        self._values_batch = []

    def run(self):
        while not self.is_interrupted() or not self._values_queue.empty():
            try:
                sleep_start = time.time()
                self._values_batch.append(self._values_queue.get(timeout=max(self._sleep_time, 0)))
                self._values_queue.task_done()
                self._sleep_time -= time.time() - sleep_start
            except queue.Empty:
                self._sleep_time = 0

            if self._sleep_time <= 0 \
                    or len(self._values_batch) >= self._MAX_VALUES_BATCH_LENGTH \
                    or len([v for v in self._values_batch if
                            v.channel_type == ChannelType.IMAGE.value]) >= self._MAX_IMAGE_VALUES_BATCH_LENGTH:  # pylint:disable=line-too-long
                self._process_batch()

        self._process_batch()

    def join(self, timeout=None):
        self.interrupt()
        super(ChannelsValuesSendingThread, self).join(timeout)

    def _process_batch(self):
        send_start = time.time()
        if self._values_batch:
            self._send_values(self._values_batch)
            self._values_batch = []
        self._sleep_time = self._SLEEP_TIME - (time.time() - send_start)

    def _send_values(self, queued_channels_values):
        def channel_key(value):
            return ChannelNameWithTypeAndNamespace(
                value.channel_name,
                value.channel_type,
                value.channel_namespace
            )

        queued_grouped_by_channel = {channel: list(values)
                                     for channel, values
                                     in groupby(sorted(queued_channels_values, key=channel_key),
                                                channel_key)}
        channels_with_values = []
        # pylint: disable=protected-access
        user_channels_by_name = self._experiment._get_channels([
            (ch.channel_name, ch.channel_type)
            for ch in queued_grouped_by_channel.keys()
            if ch.channel_namespace == ChannelNamespace.USER
        ])
        system_channels_by_name = self._experiment.get_system_channels()

        for channel_key in queued_grouped_by_channel:
            if channel_key.channel_namespace == ChannelNamespace.USER:
                channel = user_channels_by_name[channel_key.channel_name]
            else:
                channel = system_channels_by_name[channel_key.channel_name]
            last_x = channel.x if hasattr(channel, 'x') and channel.x else 0
            channel_values = []
            for queued_value in queued_grouped_by_channel[channel_key]:
                x = queued_value.channel_value.x if queued_value.channel_value.x is not None else last_x + 1
                channel_values.append(ChannelValue(ts=queued_value.channel_value.ts,
                                                   x=x,
                                                   y=queued_value.channel_value.y))
                last_x = x
            channels_with_values.append(ChannelIdWithValues(channel.id, channel_values))

        try:
            # pylint:disable=protected-access
            self._experiment._send_channels_values(channels_with_values)
        except (NeptuneApiException, IOError) as e:
            _logger.warning('Failed to send channel value: %s', e)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class GitInfo(object):
    """ Holds information about a git repository"""
    def __init__(self,
                 commit_id,
                 message="",
                 author_name="",
                 author_email="",
                 commit_date="",
                 repository_dirty=True):
        """Creates a new instance of a git repository info

        Args:
            commit_id(str): commit id sha
            message(str): commit message
            author_name(str): commit author username
            author_email(str): commit author email
            commit_date(datetime.datetime): commit datetime
            repository_dirty(bool): True, if the repository has uncommitted changes
        """
        if commit_id is None:
            raise TypeError("commit_id must not be None")

        self.commit_id = commit_id
        self.message = message
        self.author_name = author_name
        self.author_email = author_email
        self.commit_date = commit_date
        self.repository_dirty = repository_dirty

    def __eq__(self, o):
        return o is not None and self.__dict__ == o.__dict__

    def __ne__(self, o):
        return not self.__eq__(o)

    def __str__(self):
        return 'GitInfo({})'.format(self.commit_id)

    def __repr__(self):
        return str(self)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from neptune import envs


class NeptuneException(Exception):
    pass


class Uninitialized(NeptuneException):
    def __init__(self):
        super(Uninitialized, self).__init__(
            "You must initialize neptune-client first. "
            "For more information, please visit: https://github.com/neptune-ml/neptune-client#initialize-neptune")


class FileNotFound(NeptuneException):
    def __init__(self, path):
        super(FileNotFound, self).__init__("File {} doesn't exist.".format(path))


class NotAFile(NeptuneException):
    def __init__(self, path):
        super(NotAFile, self).__init__("Path {} is not a file.".format(path))


class InvalidNotebookPath(NeptuneException):
    def __init__(self, path):
        super(InvalidNotebookPath, self).__init__(
            "File {} is not a valid notebook. Should end with .ipynb.".format(path))


class InvalidChannelX(NeptuneException):
    def __init__(self, x):
        super(InvalidChannelX, self).__init__(
            "Invalid channel X-coordinate: '{}'. The sequence of X-coordinates must be strictly increasing.".format(x))


class NoChannelValue(NeptuneException):
    def __init__(self):
        super(NoChannelValue, self).__init__('No channel value provided.')


class LibraryNotInstalled(NeptuneException):
    def __init__(self, library):
        super(LibraryNotInstalled, self).__init__("Library {} is not installed".format(library))


class InvalidChannelValue(NeptuneException):
    def __init__(self, expected_type, actual_type):
        super(InvalidChannelValue, self).__init__(
            'Invalid channel value type. Expected: {expected}, actual: {actual}.'.format(
                expected=expected_type, actual=actual_type))


class NoExperimentContext(NeptuneException):
    def __init__(self):
        super(NoExperimentContext, self).__init__('Unable to find current active experiment')


class MissingApiToken(NeptuneException):
    def __init__(self):
        super(MissingApiToken, self).__init__('Missing API token. Use "{}" environment '
                                              'variable or pass it as an argument'
                                              .format(envs.API_TOKEN_ENV_NAME))


class MissingProjectQualifiedName(NeptuneException):
    def __init__(self):
        super(MissingProjectQualifiedName, self).__init__('Missing project qualified name. Use "{}" environment '
                                                          'variable or pass it as an argument'
                                                          .format(envs.PROJECT_ENV_NAME))


class IncorrectProjectQualifiedName(NeptuneException):
    def __init__(self, project_qualified_name):
        super(IncorrectProjectQualifiedName, self).__init__('Incorrect project qualified name "{}". '
                                                            'Should be in format "namespace/project_name".'
                                                            .format(project_qualified_name))



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from random import randint

import pandas as pd
from mock import MagicMock

from neptune.experiments import Experiment
from neptune.model import LeaderboardEntry
from neptune.projects import Project
from tests.neptune.api_objects_factory import a_registered_project_member, \
    an_invited_project_member
from tests.neptune.project_test_fixture import some_exp_entry_dto, some_exp_entry_row
from tests.neptune.random_utils import a_string, a_string_list, a_uuid_string


class TestProject(unittest.TestCase):
    def setUp(self):
        super(TestProject, self).setUp()
        self.client = MagicMock()
        self.project = Project(client=self.client, internal_id=a_uuid_string(), namespace=a_string(), name=a_string())

    def test_get_members(self):
        # given
        member_usernames = [a_string() for _ in range(0, 2)]
        members = [a_registered_project_member(username) for username in member_usernames]

        # and
        self.client.get_project_members.return_value = members + [an_invited_project_member()]

        # when
        fetched_member_usernames = self.project.get_members()

        # then
        self.client.get_project_members.assert_called_once_with(self.project.internal_id)

        # and
        self.assertEqual(member_usernames, fetched_member_usernames)

    def test_get_experiments_with_no_params(self):
        # given
        leaderboard_entries = [MagicMock() for _ in range(0, 2)]
        self.client.get_leaderboard_entries.return_value = leaderboard_entries

        # when
        experiments = self.project.get_experiments()

        # then
        self.client.get_leaderboard_entries.assert_called_once_with(
            project=self.project,
            ids=None,
            states=None, owners=None, tags=None,
            min_running_time=None)

        # and
        expected_experiments = [Experiment(self.client, entry.id, entry.internal_id, entry.project_full_id)
                                for entry in leaderboard_entries]
        self.assertEqual(expected_experiments, experiments)

    def test_get_experiments_with_scalar_params(self):
        # given
        leaderboard_entries = [MagicMock() for _ in range(0, 2)]
        self.client.get_leaderboard_entries.return_value = leaderboard_entries

        # and
        params = dict(
            id=a_string(),
            state='succeeded', owner=a_string(), tag=a_string(),
            min_running_time=randint(1, 100))

        # when
        experiments = self.project.get_experiments(**params)

        # then
        expected_params = dict(
            project=self.project,
            ids=[params['id']],
            states=[params['state']], owners=[params['owner']], tags=[params['tag']],
            min_running_time=params['min_running_time']
        )
        self.client.get_leaderboard_entries.assert_called_once_with(**expected_params)

        # and
        expected_experiments = [Experiment(self.client, entry.id, entry.internal_id, entry.project_full_id)
                                for entry in leaderboard_entries]
        self.assertEqual(expected_experiments, experiments)

    def test_get_experiments_with_list_params(self):
        # given
        leaderboard_entries = [MagicMock() for _ in range(0, 2)]
        self.client.get_leaderboard_entries.return_value = leaderboard_entries

        # and
        params = dict(
            id=a_string_list(),
            state=['succeeded', 'failed'], owner=a_string_list(), tag=a_string_list(),
            min_running_time=randint(1, 100))

        # when
        experiments = self.project.get_experiments(**params)

        # then
        expected_params = dict(
            project=self.project,
            ids=params['id'],
            states=params['state'], owners=params['owner'], tags=params['tag'],
            min_running_time=params['min_running_time']
        )
        self.client.get_leaderboard_entries.assert_called_once_with(**expected_params)

        # and
        expected_experiments = [Experiment(self.client, entry.id, entry.internal_id, entry.project_full_id)
                                for entry in leaderboard_entries]
        self.assertEqual(expected_experiments, experiments)

    def test_get_leaderboard(self):
        # given
        self.client.get_leaderboard_entries.return_value = [LeaderboardEntry(some_exp_entry_dto)]

        # when
        leaderboard = self.project.get_leaderboard()

        # then
        self.client.get_leaderboard_entries.assert_called_once_with(
            project=self.project,
            ids=None,
            states=None, owners=None, tags=None,
            min_running_time=None)

        # and
        expected_data = {0: some_exp_entry_row}
        expected_leaderboard = pd.DataFrame.from_dict(data=expected_data, orient='index')
        expected_leaderboard = expected_leaderboard.reindex(
            # pylint: disable=protected-access
            self.project._sort_leaderboard_columns(expected_leaderboard.columns), axis='columns')

        self.assertTrue(leaderboard.equals(expected_leaderboard))

    def test_sort_leaderboard_columns(self):
        # given
        columns_in_expected_order = [
            'id', 'name', 'created', 'finished', 'owner',
            'notes', 'size', 'tags',
            'channel_abc', 'channel_def',
            'parameter_abc', 'parameter_def',
            'property_abc', 'property_def'
        ]

        # when
        # pylint: disable=protected-access
        sorted_columns = self.project._sort_leaderboard_columns(reversed(columns_in_expected_order))

        # then
        self.assertEqual(columns_in_expected_order, sorted_columns)

    def test_full_id(self):
        # expect
        self.assertEqual(self.project.namespace + '/' + self.project.name, self.project.full_id)

    def test_to_string(self):
        # expect
        self.assertEqual('Project({})'.format(self.project.full_id), str(self.project))

    def test_repr(self):
        # expect
        self.assertEqual('Project({})'.format(self.project.full_id), repr(self.project))


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import time

from bravado.exception import HTTPError

from neptune.internal.threads.neptune_thread import NeptuneThread

_logger = logging.getLogger(__name__)


class HardwareMetricReportingThread(NeptuneThread):
    def __init__(self, metric_service, metric_sending_interval_seconds):
        super(HardwareMetricReportingThread, self).__init__(is_daemon=True)
        self.__metric_service = metric_service
        self.__metric_sending_interval_seconds = metric_sending_interval_seconds

    def run(self):
        try:
            while not self.is_interrupted():
                before = time.time()

                try:
                    self.__metric_service.report_and_send(timestamp=time.time())
                except HTTPError as e:
                    _logger.debug('Unexpected HTTP error in hardware metric reporting thread: %s', e)

                reporting_duration = time.time() - before

                time.sleep(max(0, self.__metric_sending_interval_seconds - reporting_duration))
        except Exception as e:
            _logger.debug('Unexpected error in hardware metric reporting thread: %s', e)



import os

from setuptools import find_packages, setup

import versioneer


def main():
    root_dir = os.path.dirname(__file__)

    with open(os.path.join(root_dir, 'requirements.txt')) as f:
        requirements = [r.strip() for r in f]
        setup(
            name='neptune-client',
            version=versioneer.get_version(),
            description='Neptune Client',
            author='neptune.ml',
            author_email='contact@neptune.ml',
            url='https://neptune.ml/',
            long_description='Neptune Client',
            license='Apache License 2.0',
            install_requires=requirements,
            packages=find_packages(),
            cmdclass=versioneer.get_cmdclass(),
            entry_points={
                'console_scripts': [
                    'neptune = cli.main:main',
                ],
            },
            classifiers=[
                # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
                # 'Development Status :: 1 - Planning',
                # 'Development Status :: 2 - Pre-Alpha',
                # 'Development Status :: 3 - Alpha',
                'Development Status :: 4 - Beta',
                # 'Development Status :: 5 - Production/Stable',
                # 'Development Status :: 6 - Mature',
                # 'Development Status :: 7 - Inactive',
                'Environment :: Console',
                'Intended Audience :: Developers',
                'License :: OSI Approved :: Apache Software License',
                'Operating System :: POSIX',
                'Operating System :: MacOS',
                'Operating System :: Unix',
                'Operating System :: Microsoft :: Windows',
                'Programming Language :: Python',
                'Programming Language :: Python :: 2',
                'Programming Language :: Python :: 3',
                'Topic :: Software Development :: Libraries :: Python Modules',
            ]
        )


if __name__ == "__main__":
    main()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import threading

from neptune import envs, projects, experiments
from neptune.exceptions import MissingProjectQualifiedName, Uninitialized
from neptune.sessions import Session
from ._version import get_versions

__version__ = get_versions()['version']
del get_versions

session = None
project = None

__lock = threading.RLock()



def init(project_qualified_name=None, api_token=None, proxies=None):
    if project_qualified_name is None:
        project_qualified_name = os.getenv(envs.PROJECT_ENV_NAME)

    # pylint: disable=global-statement
    with __lock:
        global session, project

        session = Session(api_token=api_token, proxies=proxies)

        if project_qualified_name is None:
            raise MissingProjectQualifiedName()

        project = session.get_project(project_qualified_name)

        return project


def set_project(project_qualified_name):
    # pylint: disable=global-statement
    with __lock:
        global session, project

        if session is None:
            session = init(project_qualified_name=project_qualified_name)
        else:
            project = session.get_project(project_qualified_name)

        return project


def create_experiment(name=None,
                      description=None,
                      params=None,
                      properties=None,
                      tags=None,
                      upload_source_files=None,
                      abort_callback=None,
                      upload_stdout=True,
                      upload_stderr=True,
                      send_hardware_metrics=True,
                      run_monitoring_thread=True,
                      handle_uncaught_exceptions=True,
                      git_info=None,
                      hostname=None):
    # pylint: disable=global-statement
    global project
    if project is None:
        raise Uninitialized()

    return project.create_experiment(
        name=name,
        description=description,
        params=params,
        properties=properties,
        tags=tags,
        upload_source_files=upload_source_files,
        abort_callback=abort_callback,
        upload_stdout=upload_stdout,
        upload_stderr=upload_stderr,
        send_hardware_metrics=send_hardware_metrics,
        run_monitoring_thread=run_monitoring_thread,
        handle_uncaught_exceptions=handle_uncaught_exceptions,
        git_info=git_info,
        hostname=hostname
    )


get_experiment = experiments.get_current_experiment


def append_tag(tag):
    get_experiment().append_tag(tag)


def remove_tag(tag):
    get_experiment().remove_tag(tag)


def set_property(key, value):
    get_experiment().set_property(key, value)


def remove_property(key):
    get_experiment().remove_property(key)


def send_metric(channel_name, x, y=None, timestamp=None):
    return get_experiment().send_metric(channel_name, x, y, timestamp)


def send_text(channel_name, x, y=None, timestamp=None):
    return get_experiment().send_text(channel_name, x, y, timestamp)


def send_image(channel_name, x, y=None, name=None, description=None, timestamp=None):
    return get_experiment().send_image(channel_name, x, y, name, description, timestamp)


def send_artifact(artifact):
    return get_experiment().send_artifact(artifact)


def stop(traceback=None):
    get_experiment().stop(traceback)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from neptune.internal.hardware.constants import BYTES_IN_ONE_GB
from neptune.internal.hardware.gauges.cpu import SystemCpuUsageGauge
from neptune.internal.hardware.gauges.gauge_factory import GaugeFactory
from neptune.internal.hardware.gauges.gauge_mode import GaugeMode
from neptune.internal.hardware.gauges.gpu import GpuMemoryGauge, GpuUsageGauge
from neptune.internal.hardware.gauges.memory import SystemMemoryUsageGauge
from neptune.internal.hardware.metrics.metric import Metric, MetricResourceType
from neptune.internal.hardware.metrics.metrics_factory import MetricsFactory
from neptune.internal.hardware.resources.system_resource_info import SystemResourceInfo


class TestMetricsFactory(unittest.TestCase):
    def setUp(self):
        self.gauge_factory = GaugeFactory(GaugeMode.SYSTEM)

    def test_create_metrics_with_gpu(self):
        # given
        system_resource_info = SystemResourceInfo(
            cpu_core_count=4,
            memory_amount_bytes=16 * BYTES_IN_ONE_GB,
            gpu_card_indices=[0, 1],
            gpu_memory_amount_bytes=8 * BYTES_IN_ONE_GB
        )
        # and
        metrics_factory = MetricsFactory(self.gauge_factory, system_resource_info)

        # when
        metrics_container = metrics_factory.create_metrics_container()

        # then
        self.assertEqual(
            Metric(
                name=u'CPU - usage',
                description=u'average of all cores',
                resource_type=MetricResourceType.CPU,
                unit=u'%',
                min_value=0.0,
                max_value=100.0,
                gauges=[SystemCpuUsageGauge()]
            ),
            metrics_container.cpu_usage_metric
        )
        # and
        self.assertEqual(
            Metric(
                name=u'RAM',
                description=u'',
                resource_type=MetricResourceType.RAM,
                unit=u'GB',
                min_value=0.0,
                max_value=16.0,
                gauges=[SystemMemoryUsageGauge()]
            ),
            metrics_container.memory_metric
        )
        # and
        self.assertEqual(
            Metric(
                name=u'GPU - usage',
                description=u'2 cards',
                resource_type=MetricResourceType.GPU,
                unit=u'%',
                min_value=0.0,
                max_value=100.0,
                gauges=[GpuUsageGauge(card_index=0), GpuUsageGauge(card_index=1)]
            ),
            metrics_container.gpu_usage_metric
        )
        # and
        self.assertEqual(
            Metric(
                name=u'GPU - memory',
                description=u'2 cards',
                resource_type=MetricResourceType.GPU_RAM,
                unit=u'GB',
                min_value=0.0,
                max_value=8.0,
                gauges=[GpuMemoryGauge(card_index=0), GpuMemoryGauge(card_index=1)]
            ),
            metrics_container.gpu_memory_metric
        )

    def test_create_metrics_without_gpu(self):
        # given
        system_resource_info = SystemResourceInfo(
            cpu_core_count=4,
            memory_amount_bytes=16 * BYTES_IN_ONE_GB,
            gpu_card_indices=[],
            gpu_memory_amount_bytes=0
        )
        # and
        metrics_factory = MetricsFactory(self.gauge_factory, system_resource_info)

        # when
        metrics_container = metrics_factory.create_metrics_container()

        # then
        self.assertIsNotNone(metrics_container.cpu_usage_metric)
        self.assertIsNotNone(metrics_container.memory_metric)
        self.assertIsNone(metrics_container.gpu_usage_metric)
        self.assertIsNone(metrics_container.gpu_memory_metric)

        # and
        self.assertEqual(
            [metrics_container.cpu_usage_metric, metrics_container.memory_metric],
            metrics_container.metrics()
        )

    def test_format_fractional_cpu_core_count(self):
        # given
        system_resource_info = SystemResourceInfo(
            cpu_core_count=0.5,
            memory_amount_bytes=2 * BYTES_IN_ONE_GB,
            gpu_card_indices=[],
            gpu_memory_amount_bytes=0
        )
        # and
        metrics_factory = MetricsFactory(self.gauge_factory, system_resource_info)

        # when
        metrics_container = metrics_factory.create_metrics_container()

        # then
        self.assertEqual(
            Metric(
                name=u'CPU - usage',
                description=u'average of all cores',
                resource_type=MetricResourceType.CPU,
                unit=u'%',
                min_value=0.0,
                max_value=100.0,
                gauges=[SystemCpuUsageGauge()]
            ),
            metrics_container.cpu_usage_metric
        )


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import io
import os

from PIL import Image
import six

from neptune.exceptions import FileNotFound, InvalidChannelValue


def get_image_content(image):
    if isinstance(image, six.string_types):
        if not os.path.exists(image):
            raise FileNotFound(image)
        with open(image, 'rb') as image_file:
            return image_file.read()

    elif isinstance(image, Image.Image):
        with io.BytesIO() as image_buffer:
            image.save(image_buffer, format='PNG')
            return image_buffer.getvalue()

    raise InvalidChannelValue(expected_type='image', actual_type=type(image).__name__)
